import logging
import jpype
import jpype.imports

from it.unibo.tuprolog.core.parsing import TermParser
from it.unibo.tuprolog.core.parsing import TermReader
from it.unibo.tuprolog.core.parsing import ParseException
from it.unibo.tuprolog.core.parsing import InvalidTermTypeException

from tuprolog.core import Term
from tuprolog.core import Struct
from tuprolog.core import Constant
from tuprolog.core import Var
from tuprolog.core import Atom
from tuprolog.core import Numeric
from tuprolog.core import Integer
from tuprolog.core import Real
from tuprolog.core import Clause
from tuprolog.jvmutils import open_file, InputStream

from tuprolog.core.operators import Operator, OperatorSet, DEFAULT_OPERATORS, EMPTY_OPERATORS

from collections.abc import Iterable
from typing import Union


logging.debug("Loaded JVM classes from it.unibo.tuprolog.core.parsing.*")


def term_parser(with_default_operators: bool=True, operators: OperatorSet=None) -> TermParser:
    if operators is None:
        if with_default_operators:
            return TermParser.getWithDefaultOperators()
        else:
            return TermParser.getWithNoOperator()
    else:
        if with_default_operators:
            return TermParser.withOperators(DEFAULT_OPERATORS.plus(operators))
        else:
            return TermParser.withOperators(operators)


def term_reader(with_default_operators: bool=True, operators: OperatorSet=None) -> TermParser:
    if operators is None:
        if with_default_operators:
            return TermReader.getWithDefaultOperators()
        else:
            return TermReader.getWithNoOperator()
    else:
        if with_default_operators:
            return TermReader.withOperators(DEFAULT_OPERATORS.plus(operators))
        else:
            return TermReader.withOperators(operators)


DEFAULT_TERM_PARSER = term_parser()


DEFAULT_TERM_READER = term_reader()


def parse_term(string: str, operators: OperatorSet=None) -> Term:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseTerm(string)
    else:
        return DEFAULT_TERM_PARSER.parseTerm(string, operators)


def parse_struct(string: str, operators: OperatorSet=None) -> Struct:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseStruct(string)
    else:
        return DEFAULT_TERM_PARSER.parseStruct(string, operators)


def parse_constant(string: str, operators: OperatorSet=None) -> Constant:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseConstant(string)
    else:
        return DEFAULT_TERM_PARSER.parseConstant(string, operators)


def parse_var(string: str, operators: OperatorSet=None) -> Var:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseVar(string)
    else:
        return DEFAULT_TERM_PARSER.parseVar(string, operators)


def parse_atom(string: str, operators: OperatorSet=None) -> Atom:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseAtom(string)
    else:
        return DEFAULT_TERM_PARSER.parseAtom(string, operators)


def parse_numeric(string: str, operators: OperatorSet=None) -> Numeric:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseNumeric(string)
    else:
        return DEFAULT_TERM_PARSER.parseNumeric(string, operators)


def parse_integer(string: str, operators: OperatorSet=None) -> Integer:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseInteger(string)
    else:
        return DEFAULT_TERM_PARSER.parseInteger(string, operators)


def parse_real(string: str, operators: OperatorSet=None) -> Real:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseReal(string)
    else:
        return DEFAULT_TERM_PARSER.parseReal(string, operators)


def parse_clause(string: str, operators: OperatorSet=None) -> Clause:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseClause(string)
    else:
        return DEFAULT_TERM_PARSER.parseClause(string, operators)


def read_term(input: Union[InputStream, str], operators: OperatorSet=None) -> Term:
    input = ensure_input_steam(input)
    if operators is None:
        return DEFAULT_TERM_READER.readTerm(input)
    else:
        return DEFAULT_TERM_READER.readTerm(input, operators)


def read_terms(input: Union[InputStream, str], operators: OperatorSet=None) -> Iterable[Term]:
    input = ensure_input_steam(input)
    if operators is None:
        return DEFAULT_TERM_READER.readTerms(input)
    else:
        return DEFAULT_TERM_READER.readTerms(input, operators)



import logging
import jpype
import jpype.imports

# from tuprolog.pyutils import iterable_or_varargs
# from tuprolog.jvmutils import jiterable, jmap

from tuprolog.core import TermComparator

logging.debug("Loaded compatibility layer for JVM subtypes of " + str(TermComparator.class_.getName()))


@jpype.JImplements(TermComparator)
class AbstractTermComparator(object):
    @jpype.JOverride
    def equals(self, other):
        return self is other

    @jpype.JOverride
    def compare(self, first, second):
        raise NotImplementedError()



import logging
import jpype
import jpype.imports

# from tuprolog.pyutils import iterable_or_varargs
# from tuprolog.jvmutils import jiterable, jmap

from tuprolog.core import Formatter
from tuprolog.core import TermFormatter
from tuprolog.core.visitors import AbstractTermVisitor

logging.debug("Loaded compatibility layer for JVM subtypes of " + str(TermFormatter.class_.getName()))


@jpype.JImplements(Formatter)
class AbstractFormatter(object):
    @jpype.JOverride
    def format(self, term):
        raise NotImplementedError()


@jpype.JImplements(TermFormatter)
class AbstractTermFormatter(AbstractFormatter, AbstractTermVisitor):
    @jpype.JOverride
    def format(self, term):
        return term.accept(self)

    @jpype.JOverride
    def defaultValue(self, term):
        return super(AbstractTermVisitor).defaultValue(self, term)

    @jpype.JOverride
    def visitTerm(self, term):
        return super(AbstractTermVisitor, self).visitTerm(term)

    @jpype.JOverride
    def visitVar(self, term):
        return super(AbstractTermVisitor, self).visitVar(term)

    @jpype.JOverride
    def visitConstant(self, term):
        return super(AbstractTermVisitor, self).visitConstant(term)

    @jpype.JOverride
    def visitStruct(self, term):
        return super(AbstractTermVisitor, self).visitStruct(term)

    @jpype.JOverride
    def visitCollection(self, term):
        return super(AbstractTermVisitor, self).visitCollection(term)

    @jpype.JOverride
    def visitAtom(self, term):
        return super(AbstractTermVisitor, self).visitAtom(term)

    @jpype.JOverride
    def visitTruth(self, term):
        return super(AbstractTermVisitor, self).visitTruth(term)

    @jpype.JOverride
    def visitNumeric(self, term):
        return super(AbstractTermVisitor, self).visitNumeric(term)

    @jpype.JOverride
    def visitInteger(self, term):
        return super(AbstractTermVisitor, self).visitInteger(term)

    @jpype.JOverride
    def visitReal(self, term):
        return super(AbstractTermVisitor, self).visitReal(term)

    @jpype.JOverride
    def visitBlock(self, term):
        return super(AbstractTermVisitor, self).visitBlock(term)

    @jpype.JOverride
    def visitEmpty(self, term):
        return super(AbstractTermVisitor, self).visitEmpty(term)

    @jpype.JOverride
    def visitEmptyBlock(self, term):
        return super(AbstractTermVisitor, self).visitEmptyBlock(term)

    @jpype.JOverride
    def visitList(self, term):
        return super(AbstractTermVisitor, self).visitList(term)

    @jpype.JOverride
    def visitCons(self, term):
        return super(AbstractTermVisitor, self).visitCons(term)

    @jpype.JOverride
    def visitEmptyList(self, term):
        return super(AbstractTermVisitor, self).visitEmptyList(term)

    @jpype.JOverride
    def visitTuple(self, term):
        return super(AbstractTermVisitor, self).visitTuple(term)

    @jpype.JOverride
    def visitIndicator(self, term):
        return super(AbstractTermVisitor, self).visitIndicator(term)

    @jpype.JOverride
    def visitClause(self, term):
        return super(AbstractTermVisitor, self).visitClause(term)

    @jpype.JOverride
    def visitRule(self, term):
        return super(AbstractTermVisitor, self).visitRule(term)

    @jpype.JOverride
    def visitFact(self, term):
        return super(AbstractTermVisitor, self).visitFact(term)

    @jpype.JOverride
    def visitDirective(self, term):
        return super(AbstractTermVisitor, self).visitDirective(term)



import logging
import jpype
import jpype.imports

from it.unibo.tuprolog.core import Atom
from it.unibo.tuprolog.core import Block
from it.unibo.tuprolog.core import Clause
from it.unibo.tuprolog.core import Cons
from it.unibo.tuprolog.core import Constant
from it.unibo.tuprolog.core import Directive
from it.unibo.tuprolog.core import Empty
from it.unibo.tuprolog.core import EmptyBlock
from it.unibo.tuprolog.core import EmptyList
from it.unibo.tuprolog.core import Fact
from it.unibo.tuprolog.core import Formatter
from it.unibo.tuprolog.core import Indicator
from it.unibo.tuprolog.core import Integer
from it.unibo.tuprolog.core import List
from it.unibo.tuprolog.core import Numeric
from it.unibo.tuprolog.core import Real
from it.unibo.tuprolog.core import Recursive
from it.unibo.tuprolog.core import Rule
from it.unibo.tuprolog.core import Scope
from it.unibo.tuprolog.core import Struct
from it.unibo.tuprolog.core import Substitution
from it.unibo.tuprolog.core import Term
from it.unibo.tuprolog.core import TermComparator
from it.unibo.tuprolog.core import TermConvertible
from it.unibo.tuprolog.core import TermFormatter
from it.unibo.tuprolog.core import Terms
from it.unibo.tuprolog.core import TermVisitor
from it.unibo.tuprolog.core import Truth
from it.unibo.tuprolog.core import Tuple
from it.unibo.tuprolog.core import Var

from tuprolog.pyutils import iterable_or_varargs
from tuprolog.jvmutils import jiterable, jmap

from typing import Iterable, Union, Dict


logging.debug("Loaded JVM classes from it.unibo.tuprolog.core.*")


@jpype.JImplements(TermConvertible)
class AbstractTermConvertible(object):
    @jpype.JOverride
    def toTerm(self):
        raise NotImplementedError()


def atom(string: str) -> Atom:
    return Atom.of(string)


def block(*terms: Union[Term, Iterable[Term]]) -> Block:
    return iterable_or_varargs(terms, lambda ts: Block.of(jiterable(ts)))


def clause(head: Term=None, *body: Union[Term, Iterable[Term]]):
    return iterable_or_varargs(body, lambda bs: Clause.of(head, jiterable(bs)))


def empty_logic_list() -> EmptyList:
    return EmptyList.getInstance()


def cons(head: Term, tail: Term=None) -> Cons:
    if tail is None:
        return Cons.singleton(head)
    else: 
        return Cons.of(head, tail)


def directive(*goals: Union[Term, Iterable[Term]]) -> Directive:
    return iterable_or_varargs(goals, lambda gs: Directive.of(jiterable(gs)))


def empty_block() -> Block:
    return EmptyBlock.getInstance()


def fact(struct: Union[Term, Iterable[Term]]) -> Fact:
    return Fact.of(struct)


def indicator(name: Union[str, Term], arity: Union[int, Term]) -> Indicator:
    return Indicator.of(name, arity)


def integer(value: Union[int, str]) -> Integer:
    return Integer.of(value)


def real(value: Union[float, str]) -> Real:
    return Real.of(value)


def rule(head: Struct, *body: Union[Term, Iterable[Term]]) -> Rule:
    return iterable_or_varargs(body, lambda bs: Rule.of(head, jiterable(bs)))


def struct(functor: str, *args: Union[Term, Iterable[Term]]) -> Struct:
    return iterable_or_varargs(args, lambda xs: Struct.of(functor, jiterable(xs)))


def truth(boolean: bool) -> Truth:
    return Truth.of(boolean)


TRUE = Truth.TRUE


FALSE = Truth.FALSE


FAIL = Truth.FAIL


def logic_tuple(first: Term, second: Term, *others: Union[Term, Iterable[Term]]) -> Tuple:
    return iterable_or_varargs(others, lambda os: Tuple.of(jiterable([first, second] + list(os))))


def var(name: str) -> Var:
    return Var.of(name)


def unifier(assignments: Dict[Var, Term]={}) -> Substitution.Unifier:
    return Substitution.unifier(jmap(assignments))


def substitution(assignments: Dict[Var, Term]={}) -> Substitution:
    return Substitution.of(jmap(assignments))


def failed() -> Substitution.Fail:
    return Substitution.failed()


Term.__lt__ = lambda this, other: this.compareTo(other) < 0
Term.__gt__ = lambda this, other: this.compareTo(other) > 0
Term.__le__ = lambda this, other: this.compareTo(other) <= 0
Term.__ge__ = lambda this, other: this.compareTo(other) >= 0
Term.__getitem__ = lambda this, item, *items: this.get(item, *items)



from collections.abc import Iterable

def iterable_or_varargs(args, dispatch=lambda x: x):
    assert isinstance(args, Iterable)
    if len(args) == 1:
        item = args[0]
        if isinstance(item, Iterable):
            return dispatch(item)
        else:
            return dispatch([item])
    else:
        return dispatch(args)
    